@@ -8,7 +8,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2005 Julian Seward 
+   Copyright (C) 2000-2006 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -37,7 +37,6 @@
 #include "pub_core_libcprint.h"
 #include "pub_core_mallocfree.h"
 #include "pub_core_options.h"
-#include "pub_core_profile.h"
 #include "pub_core_tooliface.h"
 #include "valgrind.h"
 
@@ -80,6 +79,15 @@ typedef UChar UByte;
 
       bszB == pszB + 2*sizeof(SizeT) + 2*a->rz_szB
 
+   The minimum overhead per heap block for arenas used by
+   the core is:   
+
+      32-bit platforms:  2*4 + 2*4 == 16 bytes
+      64-bit platforms:  2*8 + 2*8 == 32 bytes
+
+   In both cases extra overhead may be incurred when rounding the payload
+   size up to VG_MIN_MALLOC_SZB.
+
    Furthermore, both size fields in the block have their least-significant
    bit set if the block is not in use, and unset if it is in use.
    (The bottom 3 or so bits are always free for this because of alignment.)
@@ -429,7 +437,7 @@ void ensure_mm_init ( ArenaId aid )
 {
    static Bool     client_inited = False;
    static Bool  nonclient_inited = False;
-   static SizeT client_redzone_szB = 8;   // default: be paranoid
+   static SizeT client_rz_szB = 8;     // default: be paranoid
 
    /* We use checked red zones (of various sizes) for our internal stuff,
       and an unchecked zone of arbitrary size for the client.  Of
@@ -450,23 +458,23 @@ void ensure_mm_init ( ArenaId aid )
          // redzone size with VG_(needs_malloc_replacement)() after this module
          // has done its first allocation from the client arena.
          if (VG_(needs).malloc_replacement)
-            vg_assert(client_redzone_szB == VG_(tdict).tool_client_redzone_szB);
+            vg_assert(client_rz_szB == VG_(tdict).tool_client_redzone_szB);
          return;
       }
 
       // Check and set the client arena redzone size
       if (VG_(needs).malloc_replacement) {
-         client_redzone_szB = VG_(tdict).tool_client_redzone_szB;
+         client_rz_szB = VG_(tdict).tool_client_redzone_szB;
          // 128 is no special figure, just something not too big
-         if (client_redzone_szB > 128) {
+         if (client_rz_szB > 128) {
             VG_(printf)( "\nTool error:\n"
                          "  specified redzone size is too big (%llu)\n", 
-                         (ULong)client_redzone_szB);
+                         (ULong)client_rz_szB);
             VG_(exit)(1);
          }
       }
       // Initialise the client arena
-      arena_init ( VG_AR_CLIENT,    "client",   client_redzone_szB, 1048576 );
+      arena_init ( VG_AR_CLIENT,    "client",   client_rz_szB, 1048576 );
       client_inited = True;
 
    } else {
@@ -474,13 +482,13 @@ void ensure_mm_init ( ArenaId aid )
          return;
       }
       // Initialise the non-client arenas
-      arena_init ( VG_AR_CORE,      "core",     4,       CORE_ARENA_MIN_SZB );
-      arena_init ( VG_AR_TOOL,      "tool",     4,                  1048576 );
-      arena_init ( VG_AR_SYMTAB,    "symtab",   4,                  1048576 );
-      arena_init ( VG_AR_DEMANGLE,  "demangle", 4,                    65536 );
-      arena_init ( VG_AR_EXECTXT,   "exectxt",  4,                   262144 );
-      arena_init ( VG_AR_ERRORS,    "errors",   4,                    65536 );
-      arena_init ( VG_AR_TTAUX,     "ttaux",    4,                    65536 );
+      arena_init ( VG_AR_CORE,      "core",     4,             1048576 );
+      arena_init ( VG_AR_TOOL,      "tool",     4,             1048576 );
+      arena_init ( VG_AR_SYMTAB,    "symtab",   4,             1048576 );
+      arena_init ( VG_AR_DEMANGLE,  "demangle", 4,               65536 );
+      arena_init ( VG_AR_EXECTXT,   "exectxt",  4,              262144 );
+      arena_init ( VG_AR_ERRORS,    "errors",   4,               65536 );
+      arena_init ( VG_AR_TTAUX,     "ttaux",    4,               65536 );
       nonclient_inited = True;
    }
 
@@ -572,7 +580,7 @@ Superblock* newSuperblock ( Arena* a, SizeT cszB )
       }
    }
    vg_assert(NULL != sb);
-   //zzVALGRIND_MAKE_WRITABLE(sb, cszB);
+   //zzVALGRIND_MAKE_MEM_UNDEFINED(sb, cszB);
    vg_assert(0 == (Addr)sb % VG_MIN_MALLOC_SZB);
    sb->n_payload_bytes = cszB - sizeof(Superblock);
    a->bytes_mmaped += cszB;
@@ -588,13 +596,48 @@ static
 Superblock* findSb ( Arena* a, Block* b )
 {
    Superblock* sb;
-   for (sb = a->sblocks; sb; sb = sb->next)
-      if ((Block*)&sb->payload_bytes[0] <= b
+   static UInt n_search = 0;
+   for (sb = a->sblocks; sb; sb = sb->next) {
+      if ((Block*)&sb->payload_bytes[0] <= b 
           && b < (Block*)&sb->payload_bytes[sb->n_payload_bytes])
-         return sb;
-   VG_(printf)("findSb: can't find pointer %p in arena '%s'\n", b, a->name );
-   VG_(core_panic)("findSb: VG_(arena_free)() in wrong arena?");
-   return NULL; /*NOTREACHED*/
+         break;
+   }
+   if (!sb) {
+      VG_(printf)("findSb: can't find pointer %p in arena '%s'\n", 
+                  b, a->name );
+      VG_(core_panic)("findSb: VG_(arena_free)() in wrong arena?");
+      return NULL; /*NOTREACHED*/
+   }
+
+   /* Start of performance-enhancing hack: once every 128 (chosen
+      hackily after profiling) successful searches, move the found
+      Superblock one step closer to the start of the list.  This makes
+      future searches cheaper. */
+   if ((n_search & 0x7F) == 0) {
+      /* Move sb one step closer to the start of the list. */
+      Superblock* sb0 = a->sblocks;
+      Superblock* sb1 = NULL;
+      Superblock* sb2 = NULL;
+      Superblock* tmp;
+      while (True) {
+         if (sb0 == NULL) break;
+         if (sb0 == sb) break;
+         sb2 = sb1;
+         sb1 = sb0;
+         sb0 = sb0->next;
+      }
+      if (sb0 == sb && sb0 != NULL && sb1 != NULL && sb2 != NULL) {
+         /* sb0 points to sb, sb1 to its predecessor, and sb2 to sb1's
+            predecessor.  Swap sb0 and sb1, that is, move sb0 one step
+            closer to the start of the list. */
+         tmp = sb0->next;
+         sb2->next = sb0;
+         sb0->next = sb1;
+         sb1->next = tmp;
+      }
+   }
+   /* End of performance-enhancing hack. */
+   return sb;
 }
 
 
@@ -661,7 +704,12 @@ void swizzle ( Arena* a, UInt lno )
    if (p_best == NULL) return;
 
    pn = pp = p_best;
-   for (i = 0; i < 20; i++) {
+
+   // This loop bound was 20 for a long time, but experiments showed that
+   // reducing it to 10 gave the same result in all the tests, and 5 got the
+   // same result in 85--100% of cases.  And it's called often enough to be
+   // noticeable in programs that allocated a lot.
+   for (i = 0; i < 5; i++) {
       pn = get_next_b(pn);
       pp = get_prev_b(pp);
       if (pn < p_best) p_best = pn;
@@ -866,7 +914,7 @@ void mkFreeBlock ( Arena* a, Block* b, SizeT bszB, UInt b_lno )
 {
    SizeT pszB = bszB_to_pszB(a, bszB);
    vg_assert(b_lno == pszB_to_listNo(pszB));
-   //zzVALGRIND_MAKE_WRITABLE(b, bszB);
+   //zzVALGRIND_MAKE_MEM_UNDEFINED(b, bszB);
    // Set the size fields and indicate not-in-use.
    set_bszB(b, mk_free_bszB(bszB));
 
@@ -895,7 +943,7 @@ void mkInuseBlock ( Arena* a, Block* b, SizeT bszB )
 {
    UInt i;
    vg_assert(bszB >= min_useful_bszB(a));
-   //zzVALGRIND_MAKE_WRITABLE(b, bszB);
+   //zzVALGRIND_MAKE_MEM_UNDEFINED(b, bszB);
    set_bszB(b, mk_inuse_bszB(bszB));
    set_prev_b(b, NULL);    // Take off freelist
    set_next_b(b, NULL);    // ditto
@@ -953,8 +1001,6 @@ void* VG_(arena_malloc) ( ArenaId aid, SizeT req_pszB )
    Arena*      a;
    void*       v;
 
-   VGP_PUSHCC(VgpMalloc);
-
    ensure_mm_init(aid);
    a = arenaId_to_ArenaP(aid);
 
@@ -1026,7 +1072,6 @@ void* VG_(arena_malloc) ( ArenaId aid, SizeT req_pszB )
    sanity_check_malloc_arena(aid);
 #  endif
 
-   VGP_POPCC(VgpMalloc);
    v = get_block_payload(a, b);
    vg_assert( (((Addr)v) & (VG_MIN_MALLOC_SZB-1)) == 0 );
 
@@ -1046,21 +1091,19 @@ void VG_(arena_free) ( ArenaId aid, void* ptr )
    UInt        b_listno;
    Arena*      a;
 
-   VGP_PUSHCC(VgpMalloc);
-
    ensure_mm_init(aid);
    a = arenaId_to_ArenaP(aid);
 
    if (ptr == NULL) {
-      VGP_POPCC(VgpMalloc);
       return;
    }
       
    b = get_payload_block(a, ptr);
 
-#  ifdef DEBUG_MALLOC
-   vg_assert(blockSane(a, b));
-#  endif
+   /* If this is one of V's areas, check carefully the block we're
+      getting back.  This picks up simple block-end overruns. */
+   if (aid != VG_AR_CLIENT)
+      vg_assert(blockSane(a, b));
 
    b_bszB   = get_bszB(b);
    b_pszB   = bszB_to_pszB(a, b_bszB);
@@ -1070,6 +1113,15 @@ void VG_(arena_free) ( ArenaId aid, void* ptr )
 
    a->bytes_on_loan -= b_pszB;
 
+   /* If this is one of V's areas, fill it up with junk to enhance the
+      chances of catching any later reads of it.  Note, 0xDD is
+      carefully chosen junk :-), in that: (1) 0xDDDDDDDD is an invalid
+      and non-word-aligned address on most systems, and (2) 0xDD is a
+      value which is unlikely to be generated by the new compressed
+      Vbits representation for memcheck. */
+   if (aid != VG_AR_CLIENT)
+      VG_(memset)(ptr, 0xDD, (SizeT)b_pszB);
+
    // Put this chunk back on a list somewhere.
    b_listno = pszB_to_listNo(b_pszB);
    mkFreeBlock( a, b, b_bszB, b_listno );
@@ -1125,8 +1177,6 @@ void VG_(arena_free) ( ArenaId aid, void* ptr )
 #  endif
 
    //zzVALGRIND_FREELIKE_BLOCK(ptr, 0);
-
-   VGP_POPCC(VgpMalloc);
 }
 
 
@@ -1170,8 +1220,6 @@ void* VG_(arena_memalign) ( ArenaId aid, SizeT req_alignB, SizeT req_pszB )
    SizeT  saved_bytes_on_loan;
    Arena* a;
 
-   VGP_PUSHCC(VgpMalloc);
-
    ensure_mm_init(aid);
    a = arenaId_to_ArenaP(aid);
 
@@ -1247,8 +1295,6 @@ void* VG_(arena_memalign) ( ArenaId aid, SizeT req_alignB, SizeT req_pszB )
    sanity_check_malloc_arena(aid);
 #  endif
 
-   VGP_POPCC(VgpMalloc);
-
    vg_assert( (((Addr)align_p) % req_alignB) == 0 );
 
    //zzVALGRIND_MALLOCLIKE_BLOCK(align_p, req_pszB, 0, False);
@@ -1283,8 +1329,6 @@ void* VG_(arena_calloc) ( ArenaId aid, SizeT nmemb, SizeT bytes_per_memb )
    SizeT  size;
    UChar* p;
 
-   VGP_PUSHCC(VgpMalloc);
-
    size = nmemb * bytes_per_memb;
    vg_assert(size >= nmemb && size >= bytes_per_memb);// check against overflow
 
@@ -1294,8 +1338,6 @@ void* VG_(arena_calloc) ( ArenaId aid, SizeT nmemb, SizeT bytes_per_memb )
 
    //zzVALGRIND_MALLOCLIKE_BLOCK(p, size, 0, True);
 
-   VGP_POPCC(VgpMalloc);
-   
    return p;
 }
 
@@ -1307,8 +1349,6 @@ void* VG_(arena_realloc) ( ArenaId aid, void* ptr, SizeT req_pszB )
    UChar  *p_new;
    Block* b;
 
-   VGP_PUSHCC(VgpMalloc);
-
    ensure_mm_init(aid);
    a = arenaId_to_ArenaP(aid);
 
@@ -1321,7 +1361,6 @@ void* VG_(arena_realloc) ( ArenaId aid, void* ptr, SizeT req_pszB )
    old_pszB = get_pszB(a, b);
 
    if (req_pszB <= old_pszB) {
-      VGP_POPCC(VgpMalloc);
       return ptr;
    }
 
@@ -1331,7 +1370,6 @@ void* VG_(arena_realloc) ( ArenaId aid, void* ptr, SizeT req_pszB )
 
    VG_(arena_free)(aid, ptr);
 
-   VGP_POPCC(VgpMalloc);
    return p_new;
 }
 
