@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2005 Julian Seward 
+   Copyright (C) 2000-2006 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -57,12 +57,10 @@
    way back for the moment, until we do an OS port in earnest...]
  */
 
-#include "valgrind.h"   // for VG_USERREQ__*
-#include "coregrind.h"  // for VG_USERREQ__*
-
 #include "pub_core_basics.h"
 #include "pub_core_threadstate.h"
 #include "pub_core_aspacemgr.h"
+#include "pub_core_clreq.h"         // for VG_USERREQ__*
 #include "pub_core_dispatch.h"
 #include "pub_core_errormgr.h"      // For VG_(get_n_errs_found)()
 #include "pub_core_libcbase.h"
@@ -73,7 +71,6 @@
 #include "pub_core_machine.h"
 #include "pub_core_mallocfree.h"
 #include "pub_core_options.h"
-#include "pub_core_profile.h"
 #include "pub_core_replacemalloc.h"
 #include "pub_core_scheduler.h"
 #include "pub_core_signals.h"
@@ -87,6 +84,9 @@
 #include "vki_unistd.h"
 #include "priv_sema.h"
 
+/* #include "pub_core_debuginfo.h" */   // DEBUGGING HACK ONLY
+
+
 /* ---------------------------------------------------------------------
    Types and globals for the scheduler.
    ------------------------------------------------------------------ */
@@ -96,10 +96,10 @@
 /* Defines the thread-scheduling timeslice, in terms of the number of
    basic blocks we attempt to run each thread for.  Smaller values
    give finer interleaving but much increased scheduling overheads. */
-#define SCHEDULING_QUANTUM   50000
+#define SCHEDULING_QUANTUM   100000
 
-/* If true, a fault is Valgrind-internal (ie, a bug) */
-Bool VG_(my_fault) = True;
+/* If False, a fault is Valgrind-internal (ie, a bug) */
+Bool VG_(in_generated_code) = False;
 
 /* Counts downwards in VG_(run_innerloop). */
 UInt VG_(dispatch_ctr);
@@ -157,6 +157,11 @@ HChar* name_of_sched_event ( UInt event )
       case VEX_TRC_JMP_CLIENTREQ:     return "CLIENTREQ";
       case VEX_TRC_JMP_YIELD:         return "YIELD";
       case VEX_TRC_JMP_NODECODE:      return "NODECODE";
+      case VEX_TRC_JMP_MAPFAIL:       return "MAPFAIL";
+      case VEX_TRC_JMP_NOREDIR:       return "NOREDIR";
+      case VEX_TRC_JMP_EMWARN:        return "EMWARN";
+      case VEX_TRC_JMP_TINVAL:        return "TINVAL";
+      case VG_TRC_INVARIANT_FAILED:   return "INVFAILED";
       case VG_TRC_INNER_COUNTERZERO:  return "COUNTERZERO";
       case VG_TRC_INNER_FASTMISS:     return "FASTMISS";
       case VG_TRC_FAULT_SIGNAL:       return "FAULTSIGNAL";
@@ -308,7 +313,7 @@ void VG_(vg_yield)(void)
 }
 
 
-/* Set the standard set of blocked signals, used wheneever we're not
+/* Set the standard set of blocked signals, used whenever we're not
    running a client syscall. */
 static void block_signals(ThreadId tid)
 {
@@ -330,132 +335,6 @@ static void block_signals(ThreadId tid)
    VG_(sigprocmask)(VKI_SIG_SETMASK, &mask, NULL);
 }
 
-/* Use libc setjmp/longjmp.  longjmp must not restore signal mask
-   state, but does need to pass "val" through. */
-#define SCHEDSETJMP(tid, jumped, stmt)					\
-   do {									\
-      ThreadState * volatile _qq_tst = VG_(get_ThreadState)(tid);	\
-									\
-      (jumped) = __builtin_setjmp(_qq_tst->sched_jmpbuf);               \
-      if ((jumped) == 0) {						\
-	 vg_assert(!_qq_tst->sched_jmpbuf_valid);			\
-	 _qq_tst->sched_jmpbuf_valid = True;				\
-	 stmt;								\
-      }	else if (VG_(clo_trace_sched))					\
-	 VG_(printf)("SCHEDSETJMP(line %d) tid %d, jumped=%d\n", __LINE__, tid, jumped); \
-      vg_assert(_qq_tst->sched_jmpbuf_valid);				\
-      _qq_tst->sched_jmpbuf_valid = False;				\
-   } while(0)
-
-/* Run the thread tid for a while, and return a VG_TRC_* value to the
-   scheduler indicating what happened. */
-static
-UInt run_thread_for_a_while ( ThreadId tid )
-{
-   volatile Bool jumped;
-   volatile ThreadState *tst = VG_(get_ThreadState)(tid);
-
-   volatile UInt trc = 0;
-   volatile Int  dispatch_ctr_SAVED = VG_(dispatch_ctr);
-   volatile Int  done_this_time;
-
-   /* For paranoia purposes only */
-   volatile Addr a_vex    = (Addr) & VG_(threads)[tid].arch.vex;
-   volatile Addr a_vexsh  = (Addr) & VG_(threads)[tid].arch.vex_shadow;
-   volatile Addr a_spill  = (Addr) & VG_(threads)[tid].arch.vex_spill;
-   volatile UInt sz_vex   = (UInt) sizeof VG_(threads)[tid].arch.vex;
-   volatile UInt sz_vexsh = (UInt) sizeof VG_(threads)[tid].arch.vex_shadow;
-   volatile UInt sz_spill = (UInt) sizeof VG_(threads)[tid].arch.vex_spill;
-
-   /* Paranoia */
-   vg_assert(VG_(is_valid_tid)(tid));
-   vg_assert(VG_(is_valid_tid)(tid));
-   vg_assert(VG_(is_running_thread)(tid));
-   vg_assert(!VG_(is_exiting)(tid));
-
-   /* Even more paranoia.  Check that what we have matches
-      Vex's guest state layout requirements. */
-   if (0)
-   VG_(printf)("%p %d %p %d %p %d\n",
-               (void*)a_vex, sz_vex, (void*)a_vexsh, sz_vexsh,
-               (void*)a_spill, sz_spill );
-
-   vg_assert(VG_IS_8_ALIGNED(sz_vex));
-   vg_assert(VG_IS_8_ALIGNED(sz_vexsh));
-   vg_assert(VG_IS_16_ALIGNED(sz_spill));
-
-   vg_assert(VG_IS_4_ALIGNED(a_vex));
-   vg_assert(VG_IS_4_ALIGNED(a_vexsh));
-   vg_assert(VG_IS_4_ALIGNED(a_spill));
-
-   vg_assert(sz_vex == sz_vexsh);
-   vg_assert(a_vex + sz_vex == a_vexsh);
-
-   vg_assert(sz_spill == LibVEX_N_SPILL_BYTES);
-   vg_assert(a_vex + 2 * sz_vex == a_spill);
-
-   VGP_PUSHCC(VgpRun);
-
-#  if defined(VGA_ppc32)
-   /* This is necessary due to the hacky way vex models reservations
-      on ppc.  It's really quite incorrect for each thread to have its
-      own reservation flag/address, since it's really something that
-      all threads share (that's the whole point).  But having shared
-      guest state is something we can't model with Vex.  However, as
-      per PaulM's 2.4.0ppc, the reservation is modelled using a
-      reservation flag which is cleared at each context switch.  So it
-      is indeed possible to get away with a per thread-reservation if
-      the thread's reservation is cleared before running it.
-
-      This should be abstractified and lifted out.
-   */
-   { Int i;
-     /* Clear any existing reservation.  Be paranoid and clear them all. */
-     for (i = 0; i < VG_N_THREADS; i++)
-        VG_(threads)[i].arch.vex.guest_RESVN = 0;
-   }
-
-   /* ppc guest_state vector regs must be 16byte aligned for loads/stores */
-   vg_assert(VG_IS_16_ALIGNED(VG_(threads)[tid].arch.vex.guest_VR0));
-   vg_assert(VG_IS_16_ALIGNED(VG_(threads)[tid].arch.vex_shadow.guest_VR0));
-#  endif   
-
-   /* there should be no undealt-with signals */
-   //vg_assert(VG_(threads)[tid].siginfo.si_signo == 0);
-
-   //VG_(printf)("running EIP = %p ESP=%p\n", VG_(threads)[tid].arch.m_eip, VG_(threads)[tid].arch.m_esp);
-
-   vg_assert(VG_(my_fault));
-   VG_(my_fault) = False;
-
-   SCHEDSETJMP(tid, jumped, 
-                    trc = (UInt)VG_(run_innerloop)( (void*)&tst->arch.vex ));
-
-   //nextEIP = tst->arch.m_eip;
-   //if (nextEIP >= VG_(client_end))
-   //   VG_(printf)("trc=%d jump to %p from %p\n",
-   //		  trc, nextEIP, EIP);
-   
-   VG_(my_fault) = True;
-
-   if (jumped) {
-      /* We get here if the client took a fault, which caused our
-         signal handler to longjmp. */
-      vg_assert(trc == 0);
-      trc = VG_TRC_FAULT_SIGNAL;
-      block_signals(tid);
-   } 
-
-   done_this_time = (Int)dispatch_ctr_SAVED - (Int)VG_(dispatch_ctr) - 0;
-
-   vg_assert(done_this_time >= 0);
-   bbs_done += (ULong)done_this_time;
-
-   VGP_POPCC(VgpRun);
-   return trc;
-}
-
-
 static void os_state_clear(ThreadState *tst)
 {
    tst->os_state.lwpid       = 0;
@@ -498,11 +377,11 @@ void mostly_clear_thread_record ( ThreadId tid )
 }
 
 /*                                                                             
-   Called in the child after fork.  If the parent has multiple                 
-   threads, then we've inhereted a VG_(threads) array describing them,         
-   but only the thread which called fork() is actually alive in the            
-   child.  This functions needs to clean up all those other thread             
-   structures.                                                                 
+   Called in the child after fork.  If the parent has multiple
+   threads, then we've inherited a VG_(threads) array describing them,
+   but only the thread which called fork() is actually alive in the
+   child.  This functions needs to clean up all those other thread
+   structures.
                                                                                
    Whichever tid in the parent which called fork() becomes the                 
    master_tid in the child.  That's because the only living slot in            
@@ -569,6 +448,7 @@ void VG_(scheduler_init) ( Addr clstack_end, SizeT clstack_size )
    }
 
    tid_main = VG_(alloc_ThreadState)();
+   vg_assert(tid_main == 1);
 
    VG_(threads)[tid_main].client_stack_highest_word 
       = clstack_end + 1 - sizeof(UWord);
@@ -579,6 +459,223 @@ void VG_(scheduler_init) ( Addr clstack_end, SizeT clstack_size )
 }
 
 
+/* ---------------------------------------------------------------------
+   Helpers for running translations.
+   ------------------------------------------------------------------ */
+
+/* Use gcc's built-in setjmp/longjmp.  longjmp must not restore signal
+   mask state, but does need to pass "val" through. */
+#define SCHEDSETJMP(tid, jumped, stmt)					\
+   do {									\
+      ThreadState * volatile _qq_tst = VG_(get_ThreadState)(tid);	\
+									\
+      (jumped) = __builtin_setjmp(_qq_tst->sched_jmpbuf);               \
+      if ((jumped) == 0) {						\
+	 vg_assert(!_qq_tst->sched_jmpbuf_valid);			\
+	 _qq_tst->sched_jmpbuf_valid = True;				\
+	 stmt;								\
+      }	else if (VG_(clo_trace_sched))					\
+	 VG_(printf)("SCHEDSETJMP(line %d) tid %d, jumped=%d\n",        \
+                     __LINE__, tid, jumped);                            \
+      vg_assert(_qq_tst->sched_jmpbuf_valid);				\
+      _qq_tst->sched_jmpbuf_valid = False;				\
+   } while(0)
+
+
+/* Do various guest state alignment checks prior to running a thread.
+   Specifically, check that what we have matches Vex's guest state
+   layout requirements. */
+static void do_pre_run_checks ( volatile ThreadState* tst )
+{
+   Addr a_vex    = (Addr) & tst->arch.vex;
+   Addr a_vexsh  = (Addr) & tst->arch.vex_shadow;
+   Addr a_spill  = (Addr) & tst->arch.vex_spill;
+   UInt sz_vex   = (UInt) sizeof tst->arch.vex;
+   UInt sz_vexsh = (UInt) sizeof tst->arch.vex_shadow;
+   UInt sz_spill = (UInt) sizeof tst->arch.vex_spill;
+
+   if (0)
+   VG_(printf)("%p %d %p %d %p %d\n",
+               (void*)a_vex, sz_vex, (void*)a_vexsh, sz_vexsh,
+               (void*)a_spill, sz_spill );
+
+   vg_assert(VG_IS_8_ALIGNED(sz_vex));
+   vg_assert(VG_IS_8_ALIGNED(sz_vexsh));
+   vg_assert(VG_IS_16_ALIGNED(sz_spill));
+
+   vg_assert(VG_IS_4_ALIGNED(a_vex));
+   vg_assert(VG_IS_4_ALIGNED(a_vexsh));
+   vg_assert(VG_IS_4_ALIGNED(a_spill));
+
+   vg_assert(sz_vex == sz_vexsh);
+   vg_assert(a_vex + sz_vex == a_vexsh);
+
+   vg_assert(sz_spill == LibVEX_N_SPILL_BYTES);
+   vg_assert(a_vex + 2 * sz_vex == a_spill);
+
+#  if defined(VGA_ppc32) || defined(VGA_ppc64)
+   /* ppc guest_state vector regs must be 16 byte aligned for
+      loads/stores */
+   vg_assert(VG_IS_16_ALIGNED(& tst->arch.vex.guest_VR0));
+   vg_assert(VG_IS_16_ALIGNED(& tst->arch.vex_shadow.guest_VR0));
+#  endif   
+}
+
+
+/* Run the thread tid for a while, and return a VG_TRC_* value
+   indicating why VG_(run_innerloop) stopped. */
+static UInt run_thread_for_a_while ( ThreadId tid )
+{
+   volatile Int          jumped;
+   volatile ThreadState* tst = NULL; /* stop gcc complaining */
+   volatile UInt         trc;
+   volatile Int          dispatch_ctr_SAVED;
+   volatile Int          done_this_time;
+
+   /* Paranoia */
+   vg_assert(VG_(is_valid_tid)(tid));
+   vg_assert(VG_(is_running_thread)(tid));
+   vg_assert(!VG_(is_exiting)(tid));
+
+   tst = VG_(get_ThreadState)(tid);
+   do_pre_run_checks(tst);
+   /* end Paranoia */
+
+   //if (0) {
+   //  Char buf[100];
+   //  Bool ok = VG_(get_fnname_if_entry) ( tst->arch.vex.guest_CIA,
+   //                                       buf, 100 );
+   //  if (ok) {
+   //    Addr r2actual = tst->arch.vex.guest_GPR2;
+   //    Addr r2tocptr = VG_(get_tocptr)( tst->arch.vex.guest_CIA );
+   //    if (1) VG_(printf)("R2 act 0x%016llx toc 0x%016llx  %s\n", 
+   //			  r2actual, r2tocptr, buf);
+   //    if (r2tocptr != 0) vg_assert(r2actual == r2tocptr);
+   //  }
+   //}
+
+   trc = 0;
+   dispatch_ctr_SAVED = VG_(dispatch_ctr);
+
+#  if defined(VGA_ppc32) || defined(VGA_ppc64)
+   /* This is necessary due to the hacky way vex models reservations
+      on ppc.  It's really quite incorrect for each thread to have its
+      own reservation flag/address, since it's really something that
+      all threads share (that's the whole point).  But having shared
+      guest state is something we can't model with Vex.  However, as
+      per PaulM's 2.4.0ppc, the reservation is modelled using a
+      reservation flag which is cleared at each context switch.  So it
+      is indeed possible to get away with a per thread-reservation if
+      the thread's reservation is cleared before running it.
+   */
+   /* Clear any existing reservation that this thread might have made
+      last time it was running. */
+   VG_(threads)[tid].arch.vex.guest_RESVN = 0;
+#  endif   
+
+   /* there should be no undealt-with signals */
+   //vg_assert(VG_(threads)[tid].siginfo.si_signo == 0);
+
+   vg_assert(VG_(in_generated_code) == False);
+   VG_(in_generated_code) = True;
+
+   SCHEDSETJMP(
+      tid, 
+      jumped, 
+      trc = (UInt)VG_(run_innerloop)( (void*)&tst->arch.vex,
+                                      VG_(clo_profile_flags) > 0 ? 1 : 0 )
+   );
+
+   VG_(in_generated_code) = False;
+
+   if (jumped) {
+      /* We get here if the client took a fault that caused our signal
+         handler to longjmp. */
+      vg_assert(trc == 0);
+      trc = VG_TRC_FAULT_SIGNAL;
+      block_signals(tid);
+   } 
+
+   done_this_time = (Int)dispatch_ctr_SAVED - (Int)VG_(dispatch_ctr) - 0;
+
+   vg_assert(done_this_time >= 0);
+   bbs_done += (ULong)done_this_time;
+
+   return trc;
+}
+
+
+/* Run a no-redir translation just once, and return the resulting
+   VG_TRC_* value. */
+static UInt run_noredir_translation ( Addr hcode, ThreadId tid )
+{
+   volatile Int          jumped;
+   volatile ThreadState* tst; 
+   volatile UWord        argblock[4];
+
+   /* Paranoia */
+   vg_assert(VG_(is_valid_tid)(tid));
+   vg_assert(VG_(is_running_thread)(tid));
+   vg_assert(!VG_(is_exiting)(tid));
+
+   tst = VG_(get_ThreadState)(tid);
+   do_pre_run_checks(tst);
+   /* end Paranoia */
+
+#  if defined(VGA_ppc32) || defined(VGA_ppc64)
+   /* I don't think we need to clear this thread's guest_RESVN here,
+      because we can only get here if run_thread_for_a_while() has
+      been used immediately before, on this same thread. */
+#  endif
+
+   /* There can be 3 outcomes from VG_(run_a_noredir_translation):
+
+      - a signal occurred and the sighandler longjmp'd.  Then both [2]
+        and [3] are unchanged - hence zero.
+
+      - translation ran normally, set [2] (next guest IP) and set [3]
+        to whatever [1] was beforehand, indicating a normal (boring)
+        jump to the next block.
+
+      - translation ran normally, set [2] (next guest IP) and set [3]
+        to something different from [1] beforehand, which indicates a
+        TRC_ value.
+   */
+   argblock[0] = (UWord)hcode;
+   argblock[1] = (UWord)&VG_(threads)[tid].arch.vex;
+   argblock[2] = 0; /* next guest IP is written here */
+   argblock[3] = 0; /* guest state ptr afterwards is written here */
+
+   vg_assert(VG_(in_generated_code) == False);
+   VG_(in_generated_code) = True;
+
+   SCHEDSETJMP(
+      tid, 
+      jumped, 
+      VG_(run_a_noredir_translation)( &argblock[0] )
+   );
+
+   VG_(in_generated_code) = False;
+
+   if (jumped) {
+      /* We get here if the client took a fault that caused our signal
+         handler to longjmp. */
+      vg_assert(argblock[2] == 0); /* next guest IP was not written */
+      vg_assert(argblock[3] == 0); /* trc was not written */
+      block_signals(tid);
+      return VG_TRC_FAULT_SIGNAL;
+   } else {
+      /* store away the guest program counter */
+      VG_(set_IP)( tid, argblock[2] );
+      if (argblock[3] == argblock[1])
+         /* the guest state pointer afterwards was unchanged */
+         return VG_TRC_BORING;
+      else
+         return (UInt)argblock[3];
+   }
+}
+
+
 /* ---------------------------------------------------------------------
    The scheduler proper.
    ------------------------------------------------------------------ */
@@ -593,7 +690,8 @@ static void handle_tt_miss ( ThreadId tid )
    found = VG_(search_transtab)( NULL, ip, True/*upd_fast_cache*/ );
    if (!found) {
       /* Not found; we need to request a translation. */
-      if (VG_(translate)( tid, ip, /*debug*/False, 0/*not verbose*/, bbs_done )) {
+      if (VG_(translate)( tid, ip, /*debug*/False, 0/*not verbose*/, 
+                          bbs_done, True/*allow redirection*/ )) {
 	 found = VG_(search_transtab)( NULL, ip, True ); 
          vg_assert2(found, "VG_TRC_INNER_FASTMISS: missing tt_fast entry");
       
@@ -636,6 +734,42 @@ static void handle_syscall(ThreadId tid)
    }
 }
 
+/* tid just requested a jump to the noredir version of its current
+   program counter.  So make up that translation if needed, run it,
+   and return the resulting thread return code. */
+static UInt/*trc*/ handle_noredir_jump ( ThreadId tid )
+{
+   AddrH hcode = 0;
+   Addr  ip    = VG_(get_IP)(tid);
+
+   Bool  found = VG_(search_unredir_transtab)( &hcode, ip );
+   if (!found) {
+      /* Not found; we need to request a translation. */
+      if (VG_(translate)( tid, ip, /*debug*/False, 0/*not verbose*/, bbs_done,
+                          False/*NO REDIRECTION*/ )) {
+
+         found = VG_(search_unredir_transtab)( &hcode, ip );
+         vg_assert2(found, "unredir translation missing after creation?!");
+      
+      } else {
+	 // If VG_(translate)() fails, it's because it had to throw a
+	 // signal because the client jumped to a bad address.  That
+	 // means that either a signal has been set up for delivery,
+	 // or the thread has been marked for termination.  Either
+	 // way, we just need to go back into the scheduler loop.
+         return VG_TRC_BORING;
+      }
+
+   }
+
+   vg_assert(found);
+   vg_assert(hcode != 0);
+
+   /* Otherwise run it and return the resulting VG_TRC_* value. */ 
+   return run_noredir_translation( hcode, tid );
+}
+
+
 /* 
    Run a thread until it wants to exit.
    
@@ -650,8 +784,6 @@ VgSchedReturnCode VG_(scheduler) ( ThreadId tid )
    if (VG_(clo_trace_sched))
       print_sched_event(tid, "entering VG_(scheduler)");      
 
-   VGP_PUSHCC(VgpSched);
-
    /* set the proper running signal mask */
    block_signals(tid);
    
@@ -711,7 +843,23 @@ VgSchedReturnCode VG_(scheduler) ( ThreadId tid )
 	 print_sched_event(tid, buf);
       }
 
-      switch(trc) {
+      if (trc == VEX_TRC_JMP_NOREDIR) {
+         /* If we got a request to run a no-redir version of
+            something, do so now -- handle_noredir_jump just (creates
+            and) runs that one translation.  The flip side is that the
+            noredir translation can't itself return another noredir
+            request -- that would be nonsensical.  It can, however,
+            return VG_TRC_BORING, which just means keep going as
+            normal. */
+         trc = handle_noredir_jump(tid);
+         vg_assert(trc != VEX_TRC_JMP_NOREDIR);
+      }
+
+      switch (trc) {
+      case VG_TRC_BORING:
+         /* no special event, just keep going. */
+         break;
+
       case VG_TRC_INNER_FASTMISS:
 	 vg_assert(VG_(dispatch_ctr) > 1);
 	 handle_tt_miss(tid);
@@ -776,7 +924,7 @@ VgSchedReturnCode VG_(scheduler) ( ThreadId tid )
          show = (ew < 0 || ew >= EmWarn_NUMBER)
                    ? True
                    : counts[ew]++ < 3;
-         if (show && VG_(clo_show_emwarns)) {
+         if (show && VG_(clo_show_emwarns) && !VG_(clo_xml)) {
             VG_(message)( Vg_UserMsg,
                           "Emulation warning: unsupported action:");
             VG_(message)( Vg_UserMsg, "  %s", what);
@@ -785,7 +933,31 @@ VgSchedReturnCode VG_(scheduler) ( ThreadId tid )
          break;
       }
 
+      case VEX_TRC_JMP_EMFAIL: {
+         VexEmWarn ew;
+         HChar*    what;
+         ew   = (VexEmWarn)VG_(threads)[tid].arch.vex.guest_EMWARN;
+         what = (ew < 0 || ew >= EmWarn_NUMBER)
+                   ? "unknown (?!)"
+                   : LibVEX_EmWarn_string(ew);
+         VG_(message)( Vg_UserMsg,
+                       "Emulation fatal error -- Valgrind cannot continue:");
+         VG_(message)( Vg_UserMsg, "  %s", what);
+         VG_(get_and_pp_StackTrace)( tid, VG_(clo_backtrace_size) );
+         VG_(message)(Vg_UserMsg, "");
+         VG_(message)(Vg_UserMsg, "Valgrind has to exit now.  Sorry.");
+         VG_(message)(Vg_UserMsg, "");
+         VG_(exit)(1);
+         break;
+      }
+
+      case VEX_TRC_JMP_TRAP:
+         VG_(synth_sigtrap)(tid);
+         break;
+
       case VEX_TRC_JMP_NODECODE:
+   VG_(message)(Vg_UserMsg,
+      "valgrind: Unrecognised instruction at address %p.", VG_(get_IP)(tid));
 #define M(a) VG_(message)(Vg_UserMsg, a);
    M("Your program just tried to execute an instruction that Valgrind" );
    M("did not recognise.  There are two possible reasons for this."    );
@@ -794,7 +966,7 @@ VgSchedReturnCode VG_(scheduler) ( ThreadId tid )
    M("   warning about a bad jump, it's probably your program's fault.");
    M("2. The instruction is legitimate but Valgrind doesn't handle it,");
    M("   i.e. it's Valgrind's fault.  If you think this is the case or");
-   M("   you are not sure, please let us know."                        );
+   M("   you are not sure, please let us know and we'll try to fix it.");
    M("Either way, Valgrind will now raise a SIGILL signal which will"  );
    M("probably kill your program."                                     );
 #undef M
@@ -855,8 +1027,6 @@ VgSchedReturnCode VG_(scheduler) ( ThreadId tid )
 
    vg_assert(VG_(is_exiting)(tid));
 
-   VGP_POPCC(VgpSched);
-
    //if (VG_(clo_model_pthreads))
    //   VG_(tm_thread_exit)(tid);
    
@@ -901,7 +1071,7 @@ void VG_(nuke_all_threads_except) ( ThreadId me, VgSchedReturnCode src )
 #elif defined(VGA_amd64)
 #  define VG_CLREQ_ARGS       guest_RAX
 #  define VG_CLREQ_RET        guest_RDX
-#elif defined(VGA_ppc32)
+#elif defined(VGA_ppc32) || defined(VGA_ppc64)
 #  define VG_CLREQ_ARGS       guest_GPR4
 #  define VG_CLREQ_RET        guest_GPR3
 #else
@@ -927,6 +1097,7 @@ void VG_(nuke_all_threads_except) ( ThreadId me, VgSchedReturnCode src )
                   zztid, O_CLREQ_RET, sizeof(UWord), f); \
    } while (0)
 
+
 /* ---------------------------------------------------------------------
    Handle client requests.
    ------------------------------------------------------------------ */
@@ -1011,19 +1182,19 @@ void do_client_request ( ThreadId tid )
          break;
 
       case VG_USERREQ__PRINTF: {
-         int count = 
+         Int count = 
             VG_(vmessage)( Vg_ClientMsg, (char *)arg[1], (void*)arg[2] );
             SET_CLREQ_RETVAL( tid, count );
          break; }
 
       case VG_USERREQ__INTERNAL_PRINTF: {
-         int count = 
+         Int count = 
             VG_(vmessage)( Vg_DebugMsg, (char *)arg[1], (void*)arg[2] );
             SET_CLREQ_RETVAL( tid, count );
          break; }
 
       case VG_USERREQ__PRINTF_BACKTRACE: {
-         int count =
+         Int count =
             VG_(vmessage)( Vg_ClientMsg, (char *)arg[1], (void*)arg[2] );
             VG_(get_and_pp_StackTrace)( tid, VG_(clo_backtrace_size) );
             SET_CLREQ_RETVAL( tid, count );
@@ -1148,8 +1319,6 @@ void VG_(sanity_check_general) ( Bool force_expensive )
 {
    ThreadId tid;
 
-   VGP_PUSHCC(VgpCoreCheapSanity);
-
    if (VG_(clo_sanity_level) < 1) return;
 
    /* --- First do all the tests that we can do quickly. ---*/
@@ -1161,9 +1330,7 @@ void VG_(sanity_check_general) ( Bool force_expensive )
    /* Check that nobody has spuriously claimed that the first or
       last 16 pages of memory have become accessible [...] */
    if (VG_(needs).sanity_checks) {
-      VGP_PUSHCC(VgpToolCheapSanity);
       vg_assert(VG_TDICT_CALL(tool_cheap_sanity_check));
-      VGP_POPCC(VgpToolCheapSanity);
    }
 
    /* --- Now some more expensive checks. ---*/
@@ -1173,13 +1340,10 @@ void VG_(sanity_check_general) ( Bool force_expensive )
      || VG_(clo_sanity_level) > 1
      || (VG_(clo_sanity_level) == 1 && (sanity_fast_count % 25) == 0)) {
 
-      VGP_PUSHCC(VgpCoreExpensiveSanity);
       sanity_slow_count++;
 
       if (VG_(needs).sanity_checks) {
-          VGP_PUSHCC(VgpToolExpensiveSanity);
           vg_assert(VG_TDICT_CALL(tool_expensive_sanity_check));
-          VGP_POPCC(VgpToolExpensiveSanity);
       }
 
       /* Look for stack overruns.  Visit all threads. */
@@ -1202,20 +1366,15 @@ void VG_(sanity_check_general) ( Bool force_expensive )
                          "of running out of stack!",
 		         tid, remains);
       }
-
-      VGP_POPCC(VgpCoreExpensiveSanity);
    }
 
    if (VG_(clo_sanity_level) > 1) {
-      VGP_PUSHCC(VgpCoreExpensiveSanity);
       /* Check sanity of the low-level memory manager.  Note that bugs
          in the client's code can cause this to fail, so we don't do
          this check unless specially asked for.  And because it's
          potentially very expensive. */
       VG_(sanity_check_malloc_all)();
-      VGP_POPCC(VgpCoreExpensiveSanity);
    }
-   VGP_POPCC(VgpCoreCheapSanity);
 }
 
 /*--------------------------------------------------------------------*/
